#!/usr/bin/perl -w
# get packages summary and description from the file ARCHIVES
# for use with fou4s
# (c) Lars Ellenberg <l.g.e@web.de> Sep 2002

($PROG=$0) =~ s|^.*/||;
@ARGV == 1 or die "USAGE: $PROG package-base-name\n";

$ENV{ARCHIVES} ||= "./ARCHIVES";
$ARCHIVES=$ENV{ARCHIVES};

-e "$ARCHIVES" or die "$ARCHIVES not found\n";
-e "$ARCHIVES.idx" or do {
	# the shell code below is ugly enough without additional perl quoting
	# so export ARCHIVES ...
	$ENV{ARCHIVES} = $ARCHIVES;
	print STDERR "creating the index ... ";
	system(q{
		grep -b '^---> .*\.rpm *$' "$ARCHIVES" \
		| grep -v '\.src\.' \
		| sed 's/:---> .*\/\(.*\)\.rpm/'$'\t''\1/' \
		| sort -k 2,2 \
		> "$ARCHIVES.idx"
	}) == 0 || die "failed!\n";
	print STDERR "done\n";
};

# read the index in one gulp
open IDX,"$ARCHIVES.idx" or die "open $ARCHIVES.idx: $!\n";
undef $/; $idx=<IDX>; close IDX; $/="\n";

study $idx; # this might speed up things

$idx =~ /^(\d+)\t\Q$ARGV[0]\E$/m 
or	# not in one line to avoid backtracing
$idx =~ /^(\d+)\t\Q$ARGV[0]\E-[^-]+-[^-]+$/m;
$POS=$1;

die "'$ARGV[0]' not found in $ARCHIVES.idx\n" unless $POS;

open A,"$ARCHIVES" or die "cannot open $ARCHIVES: $!\n";
# if you know how to seek in bash, feel free to do it!
seek(A,$POS,0) or die "seek failed: $!\n";
$in_description = 0;
$group=""; $summary=""; $description="";
while(defined($_=<A>)) {
	last if /:    [dl-][rwx-]{9} /;
	$group = $1   if /:    Group       :\s*(.*?)\s*(Source RPM:.*)?\n/
		 and not $in_description;
	$in_description=1
		if /:    (Summary     :|Description :|SuSE series:)/;
	if ($in_description) {
		next if /:    Description :$/;
		last if /:    Authors:/;
		$summary = $1, next if /:    Summary     :\s*(.*?)\n/;
		s/^.*?:    (.*?)\s*$/$1\n/;
		$description .= $_;
	}
}
print "$group\n$summary\n$description"

__END__

# instead of grepping the pos again, we could read the *.idx in as a hash,
# and then access the positions this way.
# might be of benefit, if we have many queries.

open IDX,"$ARCHIVES.idx" or die "failed to open $ARCHIVES.idx: $!\n";
while(defined($_=<IDX>)) {
	if (/^(\d+)\t(\S+)$/) {
		$idx{$2}=$1;
	}
	else {
		warn "$ARCHIVES.idx:$.:unexpected line format $_";
	}
}
