#!/bin/bash
#
# SVN Info:
# $Author: markus $
# $Date: 2005-02-14 23:36:33 +0100 (Mon, 14 Feb 2005) $
# $Rev: 167 $
#
# Patch description generator by Lars Ellenberg <l.g.e@web.de>
# see list_suse_updates by the same author, and
# fou4s by Markus Gaugusch, see http://fou4s.gaugusch.at/
# or probably even yast2 online update, if you can tell it to look
# at those patch descriptions in the first place.
#
# Written in bash, not perl, since Markus likes bash better :)
#
# gpd uses rsync to generate the recursive directory listings, so this
# obviously only works with rsync enabled servers...
#

function print_patch_description ()
{
	# purpose of the update might be in the *.info or .link/*.en files
	# which I try to get into the LONGDESC

	cat <<-__EOP__
	##
	## Patch description of patch $name
	##

	Kind: ${KIND:-generated}

	Shortdescription.english: ${SUMMARY:-no summary available}

	Longdescription.english:
	This update is available from
	$RSYNCSERVER/
	$path

	This patch description was generated automaticaly from recursive
	directory listings by the fou4s patch description generator.

	Comments, requests for features, and bug reports, go to
	fou4s-devel@gaugusch.at

	${LONGDESC:-No further information available}
	Hsilgne.noitpircsedgnol:

	Size: $((size/1024))

	## file timestamp on ftp server: $date $time
	Buildtime: $unix_time

	MinYaST1Version:
	MinYaST2Version:

	Packages:
	##
	## -----> $name <-----
	##
	Filename: $name.rpm
	Label: $SUMMARY
	Series: $ARCH
	InstPath: $full_path
	Size: 0 $size
	Buildtime: $unix_time
	BuiltFrom:
	Flag:
	RpmGroup:${GROUP:+ $GROUP}
	Copyright:
	Version: $version
	StartCommand:
	Obsoletes:
	Requires:
	Provides:
	Segakcap:
	__EOP__
}

# expects path in the environment, splits it up into
# name, ver, rel, version=$ver-$rel
# -lge: path remains unchanged for now
function extract_version ()
{
	name=${path##*/}
	# path=${path%/*}
	rel=${name##*-}
	[ "$rel" == "$name" ] && return 1 # no '-' found
	rel=${rel%.rpm}  # strip rpm extension
	rel=${rel%.*}    # strip arch extension, if any
	name=${name%-*}
	ver=${name##*-}
	[ "$ver" == "$name" ] && return 1 # no second '-' found
	name=${name%-*}
	version=$ver-$rel
}

function grep_most_recent_versions() {
	# in case you wonder:
	# '$'\\\[tn]'' because older sed do not support \t\n in
	# replacement text...
	sed -ne '
	/^-/!d          ; # delete if not plain file
	/\.rpm$/!d      ; # or not rpm
	/\.src\.rpm$/d  ; # or .src.rpm
	/\.nosrc\.rpm$/d; # or .nosrc.rpm
	;# /\.x86_64\.rpm$/d; # or .x86_64.rpm
	s/^[^ ]* *//    ; # strip permissions
	s/ \+/'$'\t''/g ; # squeeze blanks into tabs
	s/.rpm$//
	;# I take it that extension (arch) is always present, and does
        ;# not contain extra dots; release of non-SuSE packages may contain
	;# dots and alphanums! this is no good, but anyways...
	;# anything/ name- version -  release     extension
	s/^\(.*\)\/\(.*-\)\([^-]\+-\)\([0-9.A-Za-z]\+\)\(\.[^-.]*\)\?$/\1|\2|\
\3\
|\4|\5.rpm/p' \
	| sed -ne '
		1~3h ; # hold it
		; # insert space separators around version numbers
		2~3{s/\([0-9]\+\)/ \1 /g;H;}
		; # and print it
		3~3{H;x;s/'$'\\\n''//g;p;}' \
	| sort -t "|" -k 4,4nr |
	# the even fields are numeric parts of the version/release
	# the odd  fields are non numeric, but should be use for
	# sorting, too, since it could be 2.05b, 0.9.7c etc... 
	# non existing fields are ignored, I truncate at 25 ;)
	  sort -t " " -s \
		-k2,2nr -k3,3r -k4,4nr -k5,5r -k6,6nr -k7,7r -k8,8nr -k9,9r \
		-k10,10nr -k11,11r -k12,12nr -k13,13r -k14,14nr -k15,15r \
		-k16,16nr -k17,17r -k18,18nr -k19,19r -k20,20nr -k21,21r -k22,22nr \
		-k23,23r -k24,24nr -k25,25r \
	| sort -t "|" -s -k 2,2 \
	| sed 's/|/\//; s/[ |]//g;'
}

# just for fun :)
_pgd_c=0
_pgd_total=0
function progressdot() {
	(( $VERBOSE )) || return
	if [ $VERBOSE -ge 2 ] ; then
		echo $*
	elif [ -t 1 ] ; then
		if [ $# -eq 0 ] ; then
			printf "%$(( 60 - _pgd_c - (_pgd_c+9)/ 10 ))d\n" $_pgd_total
			_pgd_total=0
			_pgd_c=0
		else
			if [ $_pgd_c -ge 50 ] ; then
				printf "%5d\n" $_pgd_total
				_pgd_c=0
			fi
			test ${_pgd_c: -1:1} = 0 && echo -n " "
			echo -n .
			let _pgd_c+=1
			let _pgd_total+=1
		fi
	fi
}

get_info()
# gets LONGDESC from supplementary *.info files if possible
{
	#
	# for known http/ftp urls (i.e. you provided a correct --prefix
	# we could do
	#  rpm -qip --changelog $url
	# and put this into LONGDESC
	# this transfers only 50--150KB for a several MB rpm, obviously it
	# truncates after the rpm "header" ...
	# since doing so for many packages is considerable bandwidth
	# anyways, I won't do it now.
	#  FIXME maybe introduce an option to do so.
	#
	local ret=0 in=$listing.mr out=$listing.infofiles
	rm -f $out || return
	sed -ne 's,.*\t,,;h;
	     # no useful info in here:
	     # s,^\(.*\)/\([^/]*\)-[^-]*-[^-]*$,\1/.link/\2.en,p;g;
	     s,\.patch\.rpm$,_en.info,p
	     s,\.rpm$,_en.info,p' < "$in" |
	fgrep -f - "$listing" |
	sed -ne 's,^.* ,+ /,
	     :l1;/+ ../{p;s,/$,,;s,[^/]*$,,;b l1;};' | 
	sort -u > "$out"
	if [[ -s $out ]]; then
		echo "- *" >> "$out"
		rsync $RSYNCOPTS -r --include-from="$out" \
			"$RSYNCSERVER/" "$DESTPATH/"
		ret=$?
	fi
	rm -f "$out"
	return $ret
}

function parse_listing ()
{
	local listing=$1
	local size date time full_path path name ver rel
	local version unix_time
	local hash_part rpm_info LONGDESC SUMMARY GOUP 

	if ! $RESUME && [[ -d $DESTPATH/patches ]] ; then
		rm -rf "$DESTPATH/patches.old" || {
			myError 0 "Could not rm -rf $DESTPATH/patches.old"
			exit 1
		}
		myEcho 1 "moving patches/ to patches.old/"
		mv $DESTPATH/patches{,.old} &&
		mkdir "$DESTPATH/patches" || {
			myError 0 "Could not backup and recreate patches/ dir"
			exit 1
		}
	fi

	myEcho 1 "Generating new patch descriptions to $DESTPATH"

	# only consider the most recent version per basename
	grep_most_recent_versions < "$listing" > "$listing.mr"

	# get LONGDESC from description files if available
	# if you think this is a performance hit, comment it out.
	# since there are no *.info files for most locations,
	# this is probably a noop anyways.
	# use this if you want to generate the descriptions for the
	# "normal" update tree with gpd.sh: there we have *.info files!
	cp "$listing.mr" "$listing.mr2"
	get_info

	cat "$listing.mr" |
	{ # this brace is only for the total status of progressdot :)
	while read size date time path ; do
		arch=${path%%.rpm}
		arch=${arch##*.}
		ARCH=$arch

		# strip unwanted leading slash if present
		path=${path#/}
		if [[ $path = update/$SUSEVERSION/* ]] ; then
			# NOTE that I ignore PREFIX here
			# this is for the update trees
			# should come up with something similar as the
			# descriptions in the patches{,.cont} directories
			full_path=${path#update/$SUSEVERSION/}
		else
			full_path=$PREFIX$path
		fi

		# if we cannot guess the version, ignore this file
		extract_version || continue
		test $VERBOSE -ge 3 && printf "%-40s  %-16s %s\n" $name $ver-$rel ${path%/*}

		# if we have an old p15n with same version and instpath,
		# keep it.
		oldp15n="$DESTPATH/patches.old/$name-$arch-$DESC_SERIAL"
		if [[ -e $oldp15n ]] &&
		   grep -q "^Version: $version\$" "$oldp15n" &&
		   grep -q "^InstPath: $full_path\$" "$oldp15n"
		then
			mv "$oldp15n" "$DESTPATH/patches/"
			progressdot Skip existing $name-$arch-$DESC_SERIAL
			continue
		fi

		# since we reached this, we have to create a new p15n

		# is --utc a correct assumption?
		# it is a rough guess, anyways...
		unix_time=`date --utc -d "$date $time" +%s`
		LONGDESC=''
		SUMMARY=''
		GROUP=''
		KIND=''
		rpm_info=''

		# try to get some info about this update
		# ??? are there other variants of info files ???
		#
		for info in "${path%.patch.rpm}_en.info" "${path%.rpm}_en.info" 
			# no useful info in here:
			# "${path%/*}/.link/$name.en"
		do
			break
			info=$DESTPATH/$info
			[[ -r $info ]] || continue
			LONGDESC=$(<$info)
			# test -n "$LONGDESC" && echo longdesc "$LONGDESC"
			break
		done

		# try to get some more info from the ARCHIVES or rpmdb
		IFS=$'\n'
		if [[ -e $ARCHIVES ]] ; then
			# echo calling get_info with $name
			set -- $($mypath/get_info_from_ARCHIVES.pl $name 2>/dev/null)
		else
			set -- $(rpm -q --queryformat \
				 "%{GROUP}\n%{SUMMARY}\n%{ARCH}\n%{DESCRIPTION}" \
				 $name 2>/dev/null)
		fi
		if [[ $# != 0 ]] ; then
			GROUP=$1
			SUMMARY=$2
			# ARCH=$3
			: ${LONGDESC:="${*: 4}"}
		fi
		IFS=$' \t\n'

		case "$LONGDESC" in
			# TODO add all kind of patterns here
			*[Ss]ecurity\ update*) KIND=security;;
			*[Ss]ecurity:\ Yes*) KIND=security;;
		esac
		# echo $name $version $full_path $size $unix_time $SUMMARY $GROUP
		progressdot Generating $name-$arch-$DESC_SERIAL
		print_patch_description > $DESTPATH/patches/$name-$arch-$DESC_SERIAL
	done
	progressdot
	}
	[[ -d "$DESTPATH/patches.old" ]] && rm -rf "$DESTPATH/patches.old"
	# rm -f "$listing.mr"
}
shopt -s extglob

RESUME=false # to resume a previously interupted run
unset PREFIX # prefix for url in descriptions, including any slashes.
CONFIG=
DESTPATH=
GPDQUIET=false
GPDVERBOSE=0
GPDLIST=
RSYNCSERVER=
RSYNCOPTS=
RSYNC_SERVER=	# commandline override
while [ $# -gt 0 ] ; do
	case $1 in
	--)	shift; break;;
	--config) CONFIG="$1 $2"; shift;;
	--resume) RESUME=true;;
	--prefix) PREFIX=$2; shift;;
	--server) RSYNC_SERVER=$2; shift;;
	--destpath) DESTPATH=$2; shift;;
	-q|--quiet) GPDQUIET=true; [[ $2 = 1 ]] && shift ;;
	--verbose)  GPDVERBOSE=$2; shift;;
	-v)	if [[ $2 == [0-9] ]]; then
			GPDVERBOSE=$2; shift
		else
			let GPDVERBOSE+=1
		fi ;;
	-+(v))  let GPDVERBOSE=${#1}-1;;
	-v[0-9])let GPDVERBOSE=${1#-v};;
	*) break;;
	esac
	shift
done

# use an already retrieved listing and don't get it with rsync
# this test by coincidence does the right thing when $1 is not set,
# since it is not quoted
LISTING=${1-}
if [ ! -f $LISTING ] ; then
	echo "Listing file $LISTING not found!"
	exit 1
fi

# get fou4s common functions
# actually I do only need the variables
for fou4s in `dirname $0`/fou4s ./fou4s /usr/sbin/fou4s ; do
	test -x $fou4s && source $fou4s --rightsok $CONFIG --commonfuncs && break
done
VERBOSE=$GPDVERBOSE
$GPDQUIET && VERBOSE=0

# command line override for RSYNCSERVER
if [ -n "$RSYNC_SERVER" ] ; then
	RSYNCSERVER=$RSYNC_SERVER
	GPDLIST="FAKE" # fake empty content
fi
# strip trailing slashes
RSYNCSERVER=${RSYNCSERVER%%*(/)}

# markus original fou4s and my fou4s-lge differ, among other things,
# slightly in variable names...
server=${RSYNCSERVER#rsync://}
server=${server%%/*}
: ${DLPATH:=$DLBASE}
: ${SERVERPATH:=$UPDATEPATH}
test -z "$DESTPATH" && DESTPATH=${DLPATH}/$server/$SERVERPATH
for mypath in `dirname $0` . /usr/bin ; do
	test -x $mypath/get_info_from_ARCHIVES.pl && break
done

if [ -z "$GPDLIST" -a -z "$LISTING" ] ; then
	echo "Please edit $CONFIGFILE and put something into GpdList!!"
	exit 1
fi

if [[ ${PREFIX-unset} = unset ]] ; then
	PREFIX=""
	if [[ $RSYNCSERVER = */suse/$ARCH ]]; then
		PREFIX=../../
	elif [[ $GPDLIST != @(FAKE|update/$SUSEVERSION) ]]; then
		cat <<-___
		Could not guess the prefix for the URL. Please help me with
		${0##*/} --prefix something
		If you really do not want a prefix, e.g. when you use different
		config files, and this is completely outside the suse update tree:
		${0##*/} --prefix ""
		___
		exit 1
	fi
fi
# if prefix is given, I want it to include the trailing slash
[[ $PREFIX == ?*[^/] ]] && PREFIX=$PREFIX/

# you may want to include DESC_SERIAL and RSYNCSERVER into fou4s config
# if not defined, they are defined here.
# ARCHIVES=$ARCHIVESDIR/ARCHIVES # fixme put into dlpath/.../suseversion!/
# or ARCHIVES="ARCHIVES-$SUSEVERSION" ?
for a in ${ARCHIVES:=} ./ARCHIVES `dirname $0`/ARCHIVES /var/cache/fou4s/ARCHIVES ; do
	test -f $a && ARCHIVES=$a && break
done
if ! [[ -x $mypath/get_info_from_ARCHIVES.pl ]] ; then
	ARCHIVES=
elif [[ ! -f $ARCHIVES ]] ; then
	myEcho 1 "\nHINT: Copy ARCHIVES.gz to ${ARCHIVES%/*}, and gunzip it. This will make gpd.sh write better readable patch descriptions." | fmt
	ARCHIVES=
fi
DESC_SERIAL=G
myEcho 1 "Will use -$DESC_SERIAL as description serial number"
# RSYNCSERVER=rsync://ftp.gwdg.de/SuSE/ftp.suse.com/suse/$ARCH # fixme put into fou4s.conf

if [ $VERBOSE -ge 1 ] ; then
	cat << _EOF

          Warning!
			 
          Updates from this tree are NOT necessarily 
          UPDATED,  EVEN WHEN VULNERABLE TO SECURITY 
          HOLES!

_EOF
fi


# if ! mkdir -p $DLPATH/$server/$SERVERPATH/patches ; then
if [ ! -d $DESTPATH/patches ] ; then
	echo "No directory $DESTPATH/patches"
	echo "Please run/configure fou4s to create it auomatically. (fou4s -u)"
	exit 1
fi

start_time=$SECONDS
if [ -n "$LISTING" ] ; then
	echo "reading $LISTING"
	parse_listing "$LISTING"
else
	LISTING=`mktemp $DLPATH/tmp.rsync.list.XXXXXX`
	if [ ! -f "$LISTING"  ] ; then
		echo "Error creating tempfile in $DLPATH!!"
		exit 2
	fi
	# how about: trap "rm $LISTING" EXIT ... ... ?

	echo "# $ARCH $SUSEVERSION $RSYNCSERVER $DESTPATH/patches" > $LISTING
	myEcho 1 "Connecting to rsync server $RSYNCSERVER"
	for PACKAGE_PATH in $GPDLIST ; do
		PACKAGE_PATH=${PACKAGE_PATH##*(/)}  # no leading
		PACKAGE_PATH=${PACKAGE_PATH%%*(/)}/ # but one trailing slash
		PACKAGE_PATH=${PACKAGE_PATH#FAKE/}  # nothing for FAKE ...
		myEcho 1 "Getting descriptions from $PACKAGE_PATH"
		echo "# $PACKAGE_PATH" >> $LISTING
		rsync $RSYNCOPTS -r $RSYNCSERVER/$PACKAGE_PATH \
		| grep -v kernel | grep -v " zq1/" \
		| sed 's/^\(.*\) \(.*\)$/\1 '${PACKAGE_PATH//\//\\/}'\2/' >> $LISTING
		ret=${PIPESTATUS[0]}
		# seems like rsync returns 11 for server response "no such directory"
		# some of the example locations are not available for all
		# arch/releases, so just warn
		test $ret -eq 0 -o $ret -eq 11 || exit -1
		test $ret -eq 11 && echo "$PACKAGE_PATH NOT FOUND!"
	done
	end_time=$SECONDS
	myEcho 1 "Downloaded in $((end_time-start_time)) seconds"
	start_time=$SECONDS
	parse_listing "$LISTING"
	test $VERBOSE -ge 2 && echo "Kept $LISTING" || rm $LISTING
fi
end_time=$SECONDS
myEcho 1 "Generated in $((end_time-start_time)) seconds"
